// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/src/event/event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:idb_shim/idb_shim.dart' hide Event;
import 'package:meta/meta.dart';

// TODO: Let store support multiple users
/// Stores all data (rooms, users, events) somewhere.
abstract class Store {
  bool get isOpen;

  Future<void> open();
  Future<void> close();

  Future<String> getAccessToken(UserId userId);
  Future<void> setAccessToken(UserId userId, String value);

  Future<String> getSyncToken(UserId userId);
  Future<void> setSyncToken(UserId userId, String value);

  Future<JoinedRoom> getJoinedRoom(UserId userId, RoomId id);
  Future<void> setJoinedRoom(UserId userId, RoomId id, JoinedRoom room);

  Stream<JoinedRoom> getJoinedRooms(UserId userId);

  Future<Event> getEvent(RoomId roomId, EventId id);
  Stream<Event> getEvents(RoomId id);
  Future<void> addEvent(RoomId roomId, Event event);
  Future<void> addEvents(RoomId roomId, List<Event> events);
}

class IndexedDbStore implements Store {

  final String path;
  final IdbFactory _idbFactory;
  IndexedDbStore({@required this.path, @required IdbFactory idbFactory})
    : _idbFactory = idbFactory;

  var _isOpen = false;
  bool get isOpen => _isOpen;

  Database _db;

  static const accessTokenStore = 'access_tokens';
  static const syncTokenStore = 'sync_tokens';

  static const joinedRoomStore = 'joined_rooms';
  static const eventStore = 'events';

  Future<void> open() async {
    _db = await _idbFactory.open(path, version: 1,
        onUpgradeNeeded: (VersionChangeEvent event) {
          _db = event.database;

          _db.createObjectStore(accessTokenStore);
          _db.createObjectStore(syncTokenStore);
          _db.createObjectStore(joinedRoomStore);
          _db.createObjectStore(eventStore);
    });

    _isOpen = true;
  }
  Future<void> close() async {
    _db.close();
    _isOpen = false;
  }

  Future<String> getAccessToken(UserId userId) async {
    var txn = _db.transaction(accessTokenStore, "readonly");
    var store = txn.objectStore(accessTokenStore);
    String value = await store.getObject(userId.toString());
    await txn.completed;

    return value;
  }

  Future<void> setAccessToken(UserId userId, String value) async {
    var txn = _db.transaction(accessTokenStore, 'readwrite');
    var store = txn.objectStore(accessTokenStore);
    await store.put(value, userId.toString());
    await txn.completed;
  }

  Future<String> getSyncToken(UserId userId) async {
    var txn = _db.transaction(syncTokenStore, "readonly");
    var store = txn.objectStore(syncTokenStore);
    String value = await store.getObject(userId.toString());
    await txn.completed;

    return value;
  }
  Future<void> setSyncToken(UserId userId, String value) async {
    var txn = _db.transaction(syncTokenStore, 'readwrite');
    var store = txn.objectStore(syncTokenStore);
    await store.put(value, userId.toString());
    await txn.completed;
  }

  Future<JoinedRoom> getJoinedRoom(UserId userId, RoomId id) async {
    var key = userId.toString() + id.toString();

    var txn = _db.transaction(joinedRoomStore, "readonly");
    var store = txn.objectStore(joinedRoomStore);
    JoinedRoom room = JoinedRoom.fromJson(id, this, await store.getObject(key));
    await txn.completed;

    return room;
  }

  Future<void> setJoinedRoom(UserId userId, RoomId id, JoinedRoom room) async {
    var key = userId.toString() + id.toString();

    var txn = _db.transaction(joinedRoomStore, 'readwrite');
    var store = txn.objectStore(joinedRoomStore);
    await store.put(room.toJson(), key);
    await txn.completed;
  }

  Stream<JoinedRoom> getJoinedRooms(UserId userId) async* {
    var keyPrefix = userId.toString();

    var txn = _db.transaction(joinedRoomStore, "readonly");
    var store = txn.objectStore(joinedRoomStore);
    await for (CursorWithValue cv in store.openCursor()) {
      if (cv.key.toString().startsWith(keyPrefix)) {
        yield cv.value;
      }
    }
  }

  Future<Event> getEvent(RoomId roomId, EventId id) async {
    var key = roomId.toString() + id.toString();

    var txn = _db.transaction(eventStore, "readonly");
    var store = txn.objectStore(eventStore);
    Event event = Event.fromJson(roomId, await store.getObject(key));
    await txn.completed;

    return event;
  }

  Stream<Event> getEvents(RoomId roomId) async* {
    var keyPrefix = roomId.toString();

    var txn = _db.transaction(eventStore, "readonly");
    var store = txn.objectStore(eventStore);
    await for (CursorWithValue cv in store.openCursor()) {
      if (cv.key.toString().startsWith(keyPrefix)) {
        yield cv.value;
      }
    }
  }

  Future<void> addEvent(RoomId roomId, Event event) async {
    var key = roomId.toString() + event.id.toString();

    var txn = _db.transaction(eventStore, 'readwrite');
    var store = txn.objectStore(eventStore);
    await store.put(event.toJson(), key);
    await txn.completed;
  }

  Future<void> addEvents(RoomId roomId, List<Event> events) async {
    var txn = _db.transaction(eventStore, 'readwrite');
    var store = txn.objectStore(eventStore);

    for (var event in events) {
      var key = roomId.toString() + event.id.toString();
      await store.put(event.toJson(), key);
    }

    await txn.completed;
  }
}